/***************************************************************************
 *   Copyright (C) 2007 by Mateusz Dworak aka compbatant                   *
 *   compbatant@linux.pl                                                   *
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 *   This program is distributed in the hope that it will be useful,       *
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of        *
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         *
 *   GNU General Public License for more details.                          *
 *                                                                         *
 *   You should have received a copy of the GNU General Public License     *
 *   along with this program; if not, write to the                         *
 *   Free Software Foundation, Inc.,                                       *
 *   59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.             *
 ***************************************************************************/

#include "VfsBase.hpp"
#include "Exceptions.hpp"
#include "FilesEngine.hpp"

class OperationCopy
{
public:	
	
	OperationCopy(Vfs *_vfs,Vfs *_vfs2,path _source,path _destination, Status *_status, Cancelable *_cancelable)
	:source(_source),destination(_destination),status(_status),cancelable(_cancelable),vfs(_vfs),vfs2(_vfs2)
	{
		stripSize=source.string().size();
		if(vfs->getType()=="Local")
		{
			from=false;
		}
		
	}
	
	void process(File file)
	{

		string relativeSource;		
		path destinationFile;
		
		if(file.getPath()!=source)
		{
			relativeSource=file.getPath().string().substr(stripSize+1,file.getPath().string().size()-(stripSize+1));
			destinationFile=destination/(path(relativeSource));
		}
		else
		{
			destinationFile=destination;
		}
				
		if(file.isDirectory() && !file.isSymlink())
		{			
			vfs2->createDir(destinationFile);
		}
		else
		{
			if(from)
			vfs->copyFromFile(file,destinationFile,status,cancelable);
			else
			vfs2->copyToFile(file,destinationFile,status,cancelable);
		}	
	}

private:
	path destination;
	path source;	
	int stripSize;
	Vfs *vfs;
	Vfs *vfs2;
	Status *status;
	Cancelable *cancelable;
	bool from;
	
};


class OperationGetSize
{
public:	
	OperationGetSize():size(0){}

	void process(File file)
	{
		size+=file.getSize();	
	}

	double getSize() const
	{
	    return size;
	}
	
private:
	double size;	
	
};


class OperationChangeOwner
{
public:	
	
	OperationChangeOwner(Vfs *_vfs,string _user):vfs(_vfs),user(_user)
	{
	}
	
	void process(File file)
	{
		path p=file.getPath();
		vfs->changeOwner(p,user);	
	}

private:
	string user;	
	Vfs* vfs;
	
};


class OperationChangeGroup
{
public:	
	
	OperationChangeGroup(Vfs *_vfs,string _group):vfs(_vfs),group(_group)
	{
	}
	
	void process(File file)
	{
		path p=file.getPath();
		vfs->changeGroup(p,group);	
	}

private:
	string group;	
	Vfs* vfs;
	
};



class OperationRemove
{
public:	

	OperationRemove(Vfs *_vfs):vfs(_vfs){}
	void process(File file)
	{
		path p=file.getPath();
		
		if(file.isDirectory() && !file.isSymlink())
			vfs->removeDirectory(p);
		else
			vfs->removeFile(p);
	}
private:

	Vfs*vfs;
	
};


class OperationChmod
{
public:	
	
	OperationChmod(Vfs *_vfs,unsigned int _mode):vfs(_vfs),mode(_mode)
	{
	}
	
	void process(File file)
	{
		path p=file.getPath();
		vfs->chmod(p,mode);
	}

private:
	unsigned int mode;
	Vfs *vfs;
	
};





VfsBase::VfsBase(string type)
: settings(singleton<Settings>::instance())
{
	if(type=="Local")
	VfsTypeLocal=true;
	else
	VfsTypeLocal=false;
}


VfsBase::~VfsBase()
{
}


void VfsBase::copy(Vfs * vfs2, path source, path destination,Status * status, Cancelable *cancelable)
{



	File file=readFile(source);

	try	
	{
		if(getType()=="Local")
		{
		vfs2->copyToFile(file,destination,status,cancelable);
		}
		else
		{
		copyFromFile(file,destination,status,cancelable);
		}
	}
	catch(ExceptionMustRecursive &e)
	{	
		if(source!=destination)
		{
		OperationCopy operation(this,vfs2,source,destination,status,cancelable);		
		goRecursive(file,operation);   
		} 
	}		
}



path VfsBase::symlink(const path & file)
{
	/*TODO*/
	return path("");
}





void VfsBase::init(Panel * _panel,QuestionManager *_questionManager)
{
	panel=_panel;
	questionManager=_questionManager;
	panel->setLabel("","folder_small");
}

double VfsBase::getTotalSize(path & src, Cancelable *cancelable)
{
	OperationGetSize operation;
	File file=readFile(src);
    	goRecursive(file,operation,false,cancelable);
    	return operation.getSize();
}

void VfsBase::chmodRecursive(path & source, unsigned int mode)
{

	OperationChmod operation(this,mode);
	File file=readFile(source);
    	goRecursive(file,operation);

}

void VfsBase::remove(path & source)
{
	File file=readFile(source);
	path p=file.getPath();
	
	if(!file.isDirectory())
		removeFile(p);
	else
	{
		OperationRemove operation(this);
		File file=readFile(source);
    		goRecursive(file,operation,true);		
	}					
				
					
}



void VfsBase::setFileSize( vector<string>&fields, File& file )
{
/*TODO*/
	fields.push_back(FilesEngine::printSize(file.getSize()));
}

void VfsBase::setFileOwner( vector< string > & fields, File & file )
{
	fields.push_back(file.getUser());
}

void VfsBase::setFileGroup(vector< string > & fields, File & file)
{
	fields.push_back(file.getGroup());
}

void VfsBase::setFileMime( vector< string > & fields, File & file )
{
	fields.push_back(file.getMime());
}

void VfsBase::setFilePermissions( vector< string > & fields, File & file )
{
	fields.push_back(FilesEngine::permissionsToString(file.getMode()));
}

void VfsBase::setAccessTime(vector< string > & fields, File & file)
{
	fields.push_back(FilesEngine::convertTimeToString(file.getAccessTime()));
}


void VfsBase::setCreationTime(vector< string > & fields, File & file)
{
	fields.push_back(FilesEngine::convertTimeToString(file.getCreationTime()));
}

void VfsBase::setModificationTime(vector< string > & fields, File & file)
{
	fields.push_back(FilesEngine::convertTimeToString(file.getModificationTime()));
}

void VfsBase::setPath(vector< string > & fields, File & file)
{
	fields.push_back(file.getPath().string());
}

File VfsBase::readFullFile(File file)
{
	/*TODO*/
	if(!file.isDirectory())
	file.setMime(mime.getFileMime(file.getPath(),VfsTypeLocal));
	else
	file.setMime("special/directory");
	
	vector<string>fields;	

	for(fileInfoVectorIter=fileInfoVector.begin();fileInfoVectorIter!=fileInfoVector.end();++fileInfoVectorIter)
	{
		(this->*(*fileInfoVectorIter))(fields,file);
	}
	

	file.setFields(fields);
	return file;

}



void VfsBase::changeOwnerRecursive(path & source, string user)
{
	OperationChangeOwner operation(this,user);
	File file=readFile(source);
    	goRecursive(file,operation);
}

vector< string > VfsBase::getGroups()
{
	vector<string>names;
	return names;
}

vector< string > VfsBase::getUsers()
{
	vector<string>names;
	return names;
}

void VfsBase::changeGroupRecursive(path & source, string group)
{
	OperationChangeGroup operation(this,group);
	File file=readFile(source);
    	goRecursive(file,operation);
}

void VfsBase::setHeaders(vector< string > headers)
{
	vector<string>::iterator iter;
	for(iter=headers.begin();iter!=headers.end();++iter)
	{
		if(*iter=="size")
		fileInfoVector.push_back(&VfsBase::setFileSize);
		else if(*iter=="owner")
		fileInfoVector.push_back(&VfsBase::setFileOwner);
		else if(*iter=="group")
		fileInfoVector.push_back(&VfsBase::setFileGroup);
		else if(*iter=="mime")
		fileInfoVector.push_back(&VfsBase::setFileMime);
		else if(*iter=="mode")
		fileInfoVector.push_back(&VfsBase::setFilePermissions);	
		else if(*iter=="atime")
		fileInfoVector.push_back(&VfsBase::setAccessTime);	
		else if(*iter=="mtime")
		fileInfoVector.push_back(&VfsBase::setModificationTime);
		else if(*iter=="ctime")
		fileInfoVector.push_back(&VfsBase::setCreationTime);	
		else if(*iter=="path")
		fileInfoVector.push_back(&VfsBase::setPath);	
	}
}



void VfsBase::createSymlink(path &source,path &destination)
{

}




